//******************************************************************************
// glBMP.pas by Jason Allen (jra101@home.com, delphigl.cfxweb.net)
//
// License:
// --------
// This unit is freely distributable without licensing fees and is
// provided without guarantee or warrantee expressed or  implied. This unit
// is Public Domain. Feel free to use or enhance this code. You can use the
// program for any purpose you see fit.
//
//******************************************************************************
unit glBMP;

interface

uses Windows,
     OpenGL;


const
  BM = $4D42;     // Identifies a windows bitmap (as apposed to a OS/2 bitmap)
  TGA_RGB = 2;    // Identifies an uncompressed TGA in RGB format
  READ_ONLY = 0;

const GL_CLAMP_TO_EDGE                                   = $812F;

{howabout}
type
  PByteArray = ^TByteArray;
  TByteArray = array[0..32767] of Byte;

var  Buff: array[0..MAX_PATH] of Char;
bbb:string;

procedure glDeleteTextures (n: GLsizei; textures: PGLuint); stdcall; external opengl32;
procedure glGenTextures(n: GLsizei; var textures: GLuint); stdcall; external opengl32;
procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;
function gluBuild2DMipmaps(Target: GLenum; Components, Width, Height: GLint; Format, atype: GLenum; Data: Pointer): GLint; stdcall; external glu32;



type
  pFile = ^File;

  // Record for holding pixel data for a 24bpp OpenGL image
  PGLRGBTRIPLE = ^TGLRGBTRIPLE;
  tagGLRGBTRIPLE = packed record
    red : Byte;
    green : Byte;
    blue : Byte;
  end;
  TGLRGBTRIPLE = tagGLRGBTRIPLE;

  // Record for holding pixel data for a 24bpp OpenGL image
  PGLRGBQUAD = ^TGLRGBQUAD;
  tagGLRGBQUAD = packed record
    red : Byte;
    green : Byte;
    blue : Byte;
    alpha : Byte;
  end;
  TGLRGBQUAD = tagGLRGBQUAD;

  // Array to hold palette data for 8bpp and 4bpp images
  TColorTable = Array[Byte] of TRGBQUAD;

  // Holds either the palette (4, 8bpp) or bitmasks (16bpp)
  TPalette = record
    case Boolean of
      True : (Colors : TColorTable);
      False: (redMask, greenMask, blueMask : Dword);
  end;

  // Defines the type of function that is accepted for an alpha channel callback
  TAlphaFunc = function (red, green, blue : Byte) : Byte;

  // Header type for TGA images
  TTGAHEADER = packed record
    tfType : Byte;
    tfColorMapType : Byte;
    tfImageType : Byte;
    tfColorMapSpec : Array[0..4] of Byte;
    tfOrigX : Array [0..1] of Byte;
    tfOrigY : Array [0..1] of Byte;
    tfWidth : Array [0..1] of Byte;
    tfHeight : Array [0..1] of Byte;
    tfBpp : Byte;
    tfImageDes : Byte;
  end;

  TGLBMP = object
    TextureID : GLuint;         // OpenGL texture id
{!} HatAlpha  : Boolean;

    // Public procedures: descriptions are in the individual headers below
    procedure Create(filename : String);

    procedure Destroy;
    procedure DestroyGL;

    function LoadImage(filename : String) : Boolean;
    function SaveImage(filename : String) : Boolean;

    procedure InitEmpty(w, h : Integer);

    // Misc utility functions
    function Invert : Boolean;
    function FlipVert : Boolean;
    function FlipHorz : Boolean;
    function Rotate180 : Boolean;

    // OpenGL texture utility functions
{!} procedure GenTexture(pS3TC, pAni : Boolean);
    procedure Bind;
    procedure SetFilter(min, mag : GLint);
    procedure SetTextureWrap(s, t : Glint);

    function GetWidth : Integer;
    function GetHeight : Integer;
    function GetData : Pointer;

    procedure ClearMem;

    procedure SetPixel(x,y:integer;r,g,b:byte); // by Georgy Moshkin
    procedure AddPixel(x,y:integer;r,g,b:byte); // by Georgy Moshkin

  private
    width : Integer;          // Width of the bitmap
    height : Integer;         // Height of the bitmap
    colorDepth : Integer;     // Color depth of the bitmap

    rgbBits : PGLRGBQUAD;     // Holds the image in 32bpp (RGBA) format

    palette : TPalette;       // Holds the images palette (if required)
    jpgQuality : Integer;     // Sets the quality to save a JPG as

    minFilter : GLint;        // Texture minification filter
    magFilter : GLint;        // Texture magnification filter
    texWrapS : GLint;         // Texture wrapping in s direction
    texWrapT : GLint;         // Texture wrapping in t direction

    imageSize : Integer;      // Size of BMP when its loaded

    alphaFunc : TAlphaFunc;   // Alpha channel callback

    // Functions for loading from each type of image
    function LoadBMP(filename : String) : Boolean;
    function LoadTGA(filename : String) : Boolean;

    // Functions to save to each type of image
    function SaveBMP(filename : String) : Boolean;
    function SaveTGA(filename : String) : Boolean;

    // Functions for loading x bpp images
    procedure Load32(imgFile : pFile);
    procedure Load24(imgFile : pFile);
    procedure Load16(imgFile : pFile);
    procedure Load8(imgFile : pFile);
    procedure Load4(imgFile : pFile);

    // Functions used to decode 16bpp images
    function GetRightShiftCount(dwVal : DWord) : Integer;
    function GetNumSet(dwVal : DWord) : Integer;
    function Scale8(colorValue, colorBits : Integer) : Integer;
    function GetLineWidth(bpp : Integer) : Integer;

    // Utility functions
    procedure SwapRB;
    procedure ExpandTo32(image : Pointer; pad : Integer; swapRBflag : Boolean);
    procedure SetAlpha(alpha : Byte);
  end;

implementation

{ TGLBMP }

// Name       : Create
// Purpose    : Sets up the initial values for the bitmap
// Parameters : None
// Name       : Create
// Desription : Overloaded version of above creator, initializes the image
//              and loads an image from disk
// Parameters :
//  filename - Image to load
procedure TGLBMP.Create(filename : String);
begin
  Width := 0;
  Height := 0;
  ColorDepth := 0;
  jpgQuality := 75;

  // Set defualt texture generation settings
  minFilter := GL_LINEAR;
  magFilter := GL_LINEAR;
  texWrapS := GL_REPEAT;
  texWrapT := GL_REPEAT;

  LoadImage(filename);
end;

function myext(fname:string):string;
var tmp:string;
begin
 tmp:=upcase(fname[length(fname)-2])+
      upcase(fname[length(fname)-1])+
      upcase(fname[length(fname)]);
 result:=tmp;
end;

// Name       : LoadImage
// Purpose    : Loads BMP, JPG, PNG, or TGA image files into a memory array
// Parameters :
//  filename - name and path of the image to open
// Returns    : Whether the image could be loaded or not
function TGLBMP.LoadImage(filename : String): Boolean;
begin
  Result := False;

  // If an image has already been loaded into memory, free it
  ClearMem();

  // Check extension and load image using appropriate function
  if (myext(filename)='BMP') then begin
    Result := LoadBMP(filename);
    Exit;
  end;

  if (myext(filename)='TGA') then begin
    Result := LoadTGA(filename);
    Exit;
  end;



end;

// Name       : SaveImage
// Purpose    : Save's the current image as a JPG, PNG, TGA, or BMP file
// Parameters : Name and path of file to save to (if it already exists, it is
//              overwritten!)
// Returns    : Whether image could be saved succesfully
function TGLBMP.SaveImage(filename : String) : Boolean;
begin
  Result := False;

  // Check extension and save image accordingly
  if (myext(filename)='BMP') then begin
    Result := SaveBMP(filename);
    Exit;
  end;

  if (myext(filename)='TGA') then begin
    Result := SaveTGA(filename);
    Exit;
  end;

end;


// Name       : LoadBitmap
// Purpose    : Loads a windows bitmap and converts it to OpenGL format (RGB)
// Parameters :
//   filename - name and location of the file to open
// Returns    : Whether the bitmap could be loaded successfully
function TGLBMP.LoadBMP(filename : String) : Boolean;
var
  fHeader : BITMAPFILEHEADER;   // Holds the file information for the bitmap
  iHeader : BITMAPINFOHEADER;   // Holds the height, width, colordepth, etc..
                                // for the bitmap
  bmpFile : File;               // Used to load the bitmap
  bytesRead : Integer;          // Used in read operations to check for errors

  palSize : Integer;            // size of the image's palette
begin
  Result := True;
  if {FileExists(filename)}true then begin
    // "Tie" the bmpFile variable to the filename passed in
    Assign(bmpFile, filename);

    // Set file mode to read only (defined in const section)
    FileMode := READ_ONLY;

    // Opens the file (if it doesn't exist its created) and sets the size
    // of the reads to 1 byte
    Reset(bmpFile, 1);

    // Read in the bitmap file header
    BlockRead(bmpFile, fHeader, SizeOf(fHeader));

    // Check if its a windows bitmap by comparing its "magic number" to what
    // the BMP specs say it should be
    if fHeader.bfType <> BM then begin
      Result := False;
      Close(bmpFile);
      Exit;
    end;

    // Read in the bitmap info header
    BlockRead(bmpFile, iHeader, SizeOf(iHeader));

    // Make sure that the image doesn't use a bitmap core header (not supported)
    if iHeader.biSize = SizeOf(BITMAPCOREHEADER) then begin
      Result := False;
      Close(bmpFile);
      Exit;
    end;

    // Set the bitmaps width, height, and colordepth to the bitmap
    // being loaded's values
    Width := iHeader.biWidth;
    Height := Abs(iHeader.biHeight);
    ColorDepth := iHeader.biBitCount;

    // Check for compression or lack thereof
    if iHeader.biCompression <> BI_RGB then begin
      if iHeader.biCompression = BI_BITFIELDS then begin
        // Read in the red, blue, and green bitmasks, later used to
        // decode the 16bpp image's color information
        if ColorDepth = 16 then begin
          BlockRead(bmpFile, palette.redMask, SizeOf(Dword));
          BlockRead(bmpFile, palette.greenMask, SizeOf(Dword));
          BlockRead(bmpFile, palette.blueMask, SizeOf(Dword));
        end;
      end else begin
        // RLE compression is unsupported so exit
        Result := False;
        Close(bmpFile);
        Exit;
      end;
    end;

    // Read in the images palette if its required (8 and 4bpp images)
    if ColorDepth < 16 then begin
      // Size of palette in bytes
      palSize := iHeader.biClrUsed*4;

      // Read in the palette
      BlockRead(bmpFile, palette.Colors, palSize, bytesRead);
      if bytesRead <> palSize then begin
        Result := False;
        Close(bmpFile);
        Exit;
      end;
    end;

    // Jump to the bitmaps bits
    Seek(bmpFile, fHeader.bfOffBits);

    // Get the size of the image (in bytes)
    imageSize := iHeader.biSizeImage;

    // Some images have this set to zero so if they do, manually calculate it
    if imageSize = 0 then begin
      imageSize := Height * GetLineWidth(ColorDepth);
    end;

    // Load the images pixel data depending on the color depth of the image

    case ColorDepth of
      4 : Load4(@bmpFile);
      8 : Load8(@bmpFile);
      16 : Load16(@bmpFile);
      24 : Load24(@bmpFile);
      32 : Load32(@bmpFile);
      else Result := False;
    end;


    // Check to see if image was loaded
    if not Assigned(rgbBits) then begin
      Result := False;
      Close(bmpFile);
      Exit;
    end;

    // Set the alpha channel of the image to completely opaque
    SetAlpha(255);

    // Done with the file so close it
    Close(bmpFile);
  end else
    Result := False;
end;


// Name       : LoadTGA
// Desription : Loads 24 and 32bpp (with alpha channel) TGA images
// Parameters :
//  filename - name of image to load
// Returns    : Success or failure
function TGLBMP.LoadTGA(filename : String) : Boolean;
var
  tgaFile : File;           // File pointer
  tgaHeader : TTGAHEADER;   // Holds the images header
  bytesRead : Integer;      // Used for error checking
  image : PRGBTRIPLE;       // Holds the 24bpp image
begin
  Result := True;
  if {FileExists(filename)}true then begin
    // "Tie" the tgaFile variable to the filename passed in
    Assign(tgaFile, filename);

    // Set file mode to read only (defined in const section)
    FileMode := READ_ONLY;

    // Opens the file and sets the size
    // of the reads to 1 byte
    Reset(tgaFile, 1);

    // Read in the bitmap file header
    BlockRead(tgaFile, tgaHeader, SizeOf(tgaHeader));

    // Only support uncompressed images
    if (tgaHeader.tfImageType <> TGA_RGB) then begin
      Result := False;
      Close(tgaFile);
      Exit;
    end;

    // Don't support colormapped files
    if tgaHeader.tfColorMapType <> 0 then begin
      Result := False;
      Close(tgaFile);
      Exit;
    end;

    // Get the width, height, and color depth
    Width := tgaHeader.tfWidth[0] + tgaHeader.tfWidth[1] * 256;
    Height := tgaHeader.tfHeight[0] + tgaHeader.tfHeight[1] * 256;
    ColorDepth := tgaHeader.tfBpp;
    imageSize := Width*Height*(ColorDepth div 8);

    // Only support 24 or 32bpp images
    if (ColorDepth <> 32) and (ColorDepth <> 24) then begin
      Result := False;
      Close(tgaFile);
      Exit;
    end;

    // Read in the appropriate image
    case ColorDepth of
      32 :
        begin
          // Allocate memory to hold the image
          GetMem(rgbBits, imageSize);

          // Read in the image
          BlockRead(tgaFile, rgbBits^, imageSize, bytesRead);
          if bytesRead <> imageSize then begin
            Result := False;
            Close(tgaFile);
            Exit;
          end;
        end;
      24 :
        begin
          // Allocate memory to hold the 24bpp image
          GetMem(image, imageSize);

          // Read in the image
          BlockRead(tgaFile, image^, imageSize, bytesRead);
          if bytesRead <> imageSize then begin
            Result := False;
{howabout}  FreeMem(image,imageSize);
            Close(tgaFile);
            Exit;
          end;

          // Allocate memory to hold the 32bpp image
          GetMem(rgbBits, Width*Height*4);

          // Expand it from 24bpp to 32bpp and into the rgbBits buffer
          ExpandTo32(image, 0, False);

{howabout}FreeMem(image,imageSize);

          // Set the alpha channel of the image to completely opaque
          SetAlpha(255);
        end;
    end;

    // Swap the red and blue channels
    SwapRB();

  end else
    Result := False;
end;

// Name       : SaveBitmap
// Purpose    : Saves the values in the Bits variable to a windows bitmap
// Parameters :
//   filename - name and location of the file to open
// Returns    : Whether the bitmap could be saved
function TGLBMP.SaveBMP(filename : String) : Boolean;
var
  fHeader : BITMAPFILEHEADER;   // Holds the file information for the bitmap
  iHeader : BITMAPINFOHEADER;   // Holds the height, width, colordepth, etc..
                                // for the bitmap
  bmpFile : File;               // Used to load the bitmap
  length : Integer;             // Number of bytes in the bitmap
  bytesWritten : Integer;       // Used in write operations to check for errors
begin
  Result := True;

  // Check to see if there is data in the bitmap
  if Assigned(rgbBits) then begin
    // "Tie" the bmpFile to the filename
    Assign(bmpFile, filename);

    // Creates a new file (if one exists with the same name, it is overwritten)
    // and sets the amount to transfer in write operations (1 byte at a time)
    Rewrite(bmpFile, 1);

    // Calculate the number of bytes in the image
    length := Width*Height*4;

    // Set up the file header
    fHeader.bfType := BM;     // Windows bitmap
    fHeader.bfSize := SizeOf(BITMAPFILEHEADER) + SizeOf(iHeader) + length;
    fHeader.bfReserved1 := 0; // Reserved
    fHeader.bfReserved2 := 0; // Reserved

    // This sets the distance from the start of the file to the start
    // of the bitmaps color data
    fHeader.bfOffBits := SizeOf(BITMAPFILEHEADER) + SizeOf(iHeader);

    // Save the file header to file
    BlockWrite(bmpFile, fHeader, SizeOf(fHeader), bytesWritten);
    if bytesWritten <> SizeOf(fHeader) then begin
      Result := False;
      Close(bmpFile);
      Exit;
    end;

    // Clear the info header structure
    ZeroMemory(@iHeader, SizeOf(BITMAPINFOHEADER));

    // Set up the information header
    iHeader.biSize := SizeOf(BITMAPINFOHEADER);
    iHeader.biWidth := Width;           // Current width
    iHeader.biHeight := Height;         // Current height
    iHeader.biPlanes := 1;              // Number of planes, must be set to 1
    iHeader.biBitCount := 32;           // Current color depth
    iHeader.biCompression := BI_RGB;    // No compression
    iHeader.biSizeImage := length;      // Number of bytes in bitmap

    // Save the information header to file
    BlockWrite(bmpFile, iHeader, SizeOf(iHeader), bytesWritten);
    if bytesWritten <> SizeOf(iHeader) then begin
      Result := False;
      Close(bmpFile);
      Exit;
    end;

    // Swap the red and blue channels (from RGB to BGR)
    SwapRB();

    // Save the color data to file
    BlockWrite(bmpFile, rgbBits^, length, bytesWritten);
    if bytesWritten <> length then begin
      Result := False;
      Close(bmpFile);
      Exit;
    end;

    // Swap back to RGB, so as to not screw up the existing image
    SwapRB();

    // Close the file
    Close(bmpFile);
  end else
    Result := False;
end;



// Name       : SaveTGA
// Desription : Saves a TGA image to disk
// Parameters :
//  filename - name and path to save file
// Returns    : Success or failure
function TGLBMP.SaveTGA(filename : String) : Boolean;
var
  tgaHeader : TTGAHEADER; // TGA image header
  tgaFile : File;         // File pointer
  length : Integer;       // Number of bytes in image
  bytesWritten : Integer; // Error checker
begin
  Result := True;

  // Make sure there is a file to save
  if Assigned(rgbBits) then begin
    // "Tie" the filename to the filepointer
    Assign(tgaFile, filename);

    // Create a new file and open it
    Rewrite(tgaFile, 1);

    // Clear the TGA header
    ZeroMemory(@tgaHeader, SizeOf(tgaHeader));

    // Fill the structure with info for the image to be saved
    tgaHeader.tfImageType := TGA_RGB;
    tgaHeader.tfWidth[0] := Width mod 256;
    tgaHeader.tfWidth[1] := Width div 256;
    tgaHeader.tfHeight[0] := Height mod 256;
    tgaHeader.tfHeight[1] := Height div 256;
    tgaHeader.tfBpp := 32;

    // Write the header to disk
    BlockWrite(tgaFile, tgaHeader, SizeOf(tgaHeader), bytesWritten);
    if bytesWritten <> SizeOf(tgaHeader) then begin
      Result := False;
      Close(tgaFile);
      Exit;
    end;

    // Switch the red and blue channels (from RGB to BGR)
    SwapRB();

    // Calculate number of bytes in image
    length := Width*Height*4;

    // Save the image contents to file
    BlockWrite(tgaFile, rgbBits^, length, bytesWritten);
    if bytesWritten <> length then begin
      Result := False;
      Close(tgaFile);
      Exit;
    end;

    // Swap back to RGB, so as to not screw up the existing image
    SwapRB();

    Close(tgaFile);
  end else
    Result := False;
end;

procedure TGLBMP.InitEmpty(w, h : Integer);
begin
  ClearMem();
  GetMem(rgbBits, w*h*4);

  Width := w;
  Height := h;
  ColorDepth := 32;

  minFilter := GL_LINEAR;
  magFilter := GL_LINEAR;
  texWrapS := GL_REPEAT;
  texWrapT := GL_REPEAT;
end;

// Name       : Load32
// Purpose    : Loads a 32bpp bitmap image
// Parameters :
//  bmpFile : Pointer to a open bitmap file
//  length : number of bytes in the image
// Returns    : Loaded bits or nil if there was a problem
procedure TGLBMP.Load32(imgFile : pFile);
var
  bytesRead : Integer;  // Used in read operations to check for errors
  size : Integer;
begin

  size := Width*Height*4;

  // Allocate enough memory to hold the BGR image
  GetMem(rgbBits, size);


  // Read in the bitmaps color bytes
  BlockRead(imgFile^, rgbBits^, size, bytesRead);
  if bytesRead <> size then begin
    Close(imgFile^);
    Exit;
  end;

  SwapRB();
end;

// Name       : Load24
// Purpose    : Loads a 24bpp bitmap
// Parameters : Pointer to an open file, the file pointer must be pointing at
//              the bitmap data already
procedure TGLBMP.Load24(imgFile : pFile);
var
  bgrBits24 : PRGBTRIPLE;   // Temporary image buffer
  bytesRead : Integer;      // File IO error check
  pad : Integer;            // Line padding
begin
  // Calculate amount of padding at the end of each line in the image
  pad := GetLineWidth(24) - (Width*3);

  // Allocate enough memory for the temp bitmap
  GetMem(bgrBits24, imageSize);

  // Read in the bitmaps color bytes
  BlockRead(imgFile^, bgrBits24^, imageSize, bytesRead);
  if bytesRead <> imageSize then begin
    Close(imgFile^);
    Exit;
  end;

  // Allocate memory for the bitmap
  GetMem(rgbBits, Width*Height*4);

  ExpandTo32(bgrBits24, pad, True);
  FreeMem(bgrBits24,imageSize);{howabout}

  ColorDepth := 32;
end;

// Name       : Load16
// Purpose    : Loads a 16bpp image into a 32bpp memory buffer
// Parameters :
//  bmpFile - Pointer to an open file with the file pointer at the bitmap bits
//  length - Number of bytes in source image
// Returns    : A pointer to the image data (in 32bpp RGB format)
procedure TGLBMP.Load16(imgFile : pFile);
var
  bytesRead : Integer;  // Error checker
  source16 : PWord;     // Pointer to 16bpp image
  dest : PGLRGBQUAD;    // Pointer to 32bpp image
  bgrBits16 : PWord;    // 16bpp image

  x, y : Integer;       // Loop counters

  pad : Integer;      // Amount to pad end of each line in 16bpp image

  redBits : Integer;    // Number of bits set in the redmask
  greenBits : Integer;  // Number of bits set in the greenmask
  blueBits : Integer;   // Number of bits set in the blue mask

  redShr : Integer;     // Amount to shift the red channel right
  greenShr : Integer;   // Amount to shift the green channel right

begin
  // Allocate enough memory for the temp bitmap
  GetMem(bgrBits16, imageSize);

  // Read in the bitmaps color bytes
  BlockRead(imgFile^, bgrBits16^, imageSize, bytesRead);
  if bytesRead <> imageSize then begin
    Close(imgFile^);
    Exit;
  end;

  // Allocate memory to hold 32bpp image
  GetMem(rgbBits, Width*Height*4);

  // Calculate end of line padding
  pad := GetLineWidth(16) - (Width*SizeOf(Word));

  // Calculate the shift amounts (number of 0 bits before the first 1 (set) bit)
  redShr := GetRightShiftCount(palette.redMask);
  greenShr := GetRightShiftCount(palette.greenMask);

  // Calculate the number of set bits in the color mask
  redBits := GetNumSet(palette.redMask);
  greenBits := GetNumSet(palette.greenMask);
  blueBits := GetNumSet(palette.blueMask);

  source16 := bgrBits16;
  dest := rgbBits;

  // Loop through image getting color data extracting the red, green, and blue
  // channel data from the current word in the source image. This is done by
  // applying the mask to the source word and scaling the result (otherwise image
  // colors are faded)
  for x := 0 to Height-1 do begin
    for y := 0 to Width-1 do begin
      dest^.red := Scale8(source16^ and Integer(palette.redMask) shr redShr, redBits);
      dest^.green := Scale8(source16^ and Integer(palette.greenMask) shr greenShr, greenBits);
      dest^.blue := Scale8(source16^ and Integer(palette.blueMask), blueBits);
      Inc(source16);
      Inc(dest);
    end;
    source16 := Ptr(Integer(source16) + pad);
  end;
  ColorDepth := 32;
  FreeMem(bgrBits16,imageSize);{howabout}
end;

// Name       : Load8
// Purpose    : Loads a 8bpp bitmap image into a 24bpp memory buffer
// Parameters :
//  bmpFile - Pointer to an open file with the file pointer at the bitmap bits
//  length - Number of bytes in source image
// Returns    : A pointer to the image data (in 24bpp RGB format)
procedure TGLBMP.Load8(imgFile : pFile);
var
  bytesRead : Integer;  // Error checker
  source8 : PByte;      // Pointer to 8bpp image
  dest : PGLRGBQUAD;  // Pointer to 24bpp image
  bgrBits8 : PByte;     // Temp image buffer
  x, y : Integer;       // Loop counters

  pad : Integer;       // Amount to pad end of each line in 8pp image
begin
  // Allocate enough memory for the temp bitmap
  GetMem(bgrBits8, imageSize);

  // Read in the bitmaps color bytes
  BlockRead(imgFile^, bgrBits8^, imageSize, bytesRead);
  if bytesRead <> imageSize then begin
    Close(imgFile^);
    Exit;
  end;

  // Allocate memory to hold 32bpp image
  GetMem(rgbBits, Width*Height*4);

  // Calculate the amount to pad end of each line in 8pp image
  pad := GetLineWidth(8) - Width;

  source8 := bgrBits8;
  dest := rgbBits;

  // Loop through image getting color values from the palette
  for x := 0 to Height-1 do begin
    for y := 0 to Width-1 do begin
      dest^.red := palette.Colors[source8^].rgbRed;
      dest^.green := palette.Colors[source8^].rgbGreen;
      dest^.blue := palette.Colors[source8^].rgbBlue;
      Inc(dest);
      Inc(source8);
    end;
    source8 := Ptr(Integer(source8) + pad);
  end;
  ColorDepth := 32;
  FreeMem(bgrBits8,imageSize);{howabout}
end;

// Name       : Load4
// Purpose    : Loads a 4bpp image into a 24bpp memory buffer
// Parameters :
//  bmpFile - Pointer to an open file with the file pointer at the bitmap bits
//  length - Number of bytes in source image
// Returns    : A pointer to the image data (in 24bpp RGB format)
procedure TGLBMP.Load4(imgFile : pFile);
var
  bytesRead : Integer;  // Error checker
  source4 : PByte;      // Pointer to 4bpp image
  dest : PGLRGBQUAD;    // Pointer to 24bpp image
  bgrBits4 : PByte;     // 4bpp image
  x, y : Integer;       // Loop counters

  pad : Integer;        // Amount to pad end of each line in 4pp image

  palEntry : Integer;   // Which palette entry to use
begin
  // Allocate enough memory for the bitmap
  GetMem(bgrBits4, imageSize);

  // Read in the bitmaps color bytes
  BlockRead(imgFile^, bgrBits4^, imageSize, bytesRead);
  if bytesRead <> imageSize then begin
    Close(imgFile^);
    Exit;
  end;

  // Allocate memory for the bitmap
  GetMem(rgbBits, Width*Height*4);

  // Calculate amount to pad end of each line in 4bpp image
  pad := GetLineWidth(4) - (Width shr 1);

  source4 := bgrBits4;
  dest := rgbBits;

  // Loop through image reading pixel colors from the palette. Each
  // Byte in the source image is two pixels, the upper and lower nibble of
  // the byte contain an index into the palette for the image
  for x := 0 to Height-1 do begin
    for y := 0 to Width-1 do begin
      if Odd(y) then
        palEntry := Integer(source4^ and $F)
      else
        palEntry := Integer((source4^ shr 4) and $F);

      dest^.red := palette.Colors[palEntry].rgbRed;
      dest^.green := palette.Colors[palEntry].rgbGreen;
      dest^.blue := palette.Colors[palEntry].rgbBlue;


      Inc(dest);
      if Odd(y) then
        Inc(source4);
    end;
    source4 := Ptr(Integer(source4) + pad);
  end;

  ColorDepth := 32;
  FreeMem(bgrBits4,imageSize);{howabout}
end;

// Name       : Invert
// Purpose    : Inverts the color data in the current bitmap
// Parameters : None
// Returns    : Success or failure
function TGLBMP.Invert : Boolean;
var
  I : Integer;    // Loop counter
  pixel : PByte;  // pointer to current pixel
begin
  Result := True;

  // Make sure there is a bitmap to invert
  if Assigned(rgbBits) then begin
    pixel := PByte(rgbBits);

    // Loop through image, inverting the colors
    for I := 0 to Width*Height*4-1 do begin
      pixel^ := 255 - pixel^;
      Inc(pixel);
    end;
  end else
    Result := False;
end;

// Name       : FlipVert
// Purpose    : Flips the image vertically (around X axis)
// Parameters : None
// Returns    : Success or failure
function TGLBMP.FlipVert : Boolean;
var
  I : Integer;            // Loop counter

  top : PGLRGBQUAD;          // Pointer to top line in the image
  bottom : PGLRGBQUAD;       // Pointer to bottom line in the image
  tmpBits : PGLRGBQUAD; // Temporarily holds one line in image
  line : Integer;
begin
  Result := True;
  if Assigned(rgbBits) then begin
    line := Width * 4;

    // Allocate enough memory in our temp buffer to hold a whole line
    // in the image
    GetMem(tmpBits, line);

    top := rgbBits;                                  // Point to top
    bottom := Ptr(Integer(rgbBits) + line*(Height-1));   // Point to bottom

    // Loop through image and swap the top and bottom lines, then move the top
    // pointer down a row and the bottom pointer up a row and repeat
    for I := 0 to (Height shr 1)-1 do begin
      Move(top^, tmpBits^, line);
      Move(bottom^, top^, line);
      Move(tmpBits^, bottom^, line);
      top := Ptr(Integer(top) + line);
      bottom := Ptr(Integer(bottom) - line);
    end;
    FreeMem(tmpBits,line);
  end else
    Result := False;
end;

// Name       : FlipHorz
// Purpose    : Flips the image horizontally (around Y axis)
// Parameters : None
// Returns    : Success or failure
function TGLBMP.FlipHorz : Boolean;
var
  x, y : Integer;         // Loop counters

  // 32bpp image pointers
  front32 : PGLRGBQUAD;
  back32 : PGLRGBQUAD;
  tmp32 : TGLRGBQUAD;     // Temp buffer for swapping pixels

  line : Integer;
begin
  Result := True;
  if Assigned(rgbBits) then begin
    line := Width*4;
    for y := 0 to Height-1 do begin
      // Point to the front of the current line
      front32 := Ptr(Integer(rgbBits) + line*y);

      // Point to the end of the current line
      back32 := Ptr(Integer(rgbBits) + (line*y) + line-4);

      // Swap each pixel in the line
      for x := 0 to (Width shr 1)-1 do begin
        tmp32 := front32^;
        front32^ := back32^;
        back32^ := tmp32;
        Inc(front32);
        Dec(back32);
      end;
    end;
  end else
    Result := False;
end;

// Name       : Rotate180
// Purpose    : Rotates the image 180 degrees by using the FlipHorz and FlipVert
//              functions together
// Parameters : None
// Returns    : Success or failure
function TGLBMP.Rotate180 : Boolean;
begin
  Result := True;
  // First flip it horizontally
  if not (FlipHorz()) then begin
    Result := False;
    Exit;
  end;

  // Then flip it vertically
  if not (FlipVert()) then begin
    Result := False;
    Exit;
  end;
end;

// Name       : GenTexture
// Purpose    : Generates an OpenGL texture from the currently loaded image, this texture
//              can be used (applyed to a surface) by calling the Bind function before drawing
//              geometry
// Parameters : None
procedure TGLBMP.GenTexture(pS3TC, pAni : Boolean);
var
 MaxAnisotropy : GLUInt;
begin

  if Assigned(rgbBits) then begin
    // If a texture object has already been created, delete it
    if TextureID > 0 then
      glDeleteTextures(1, @TextureID);
    // Create a new OpenGL texture object
    glGenTextures(1, TextureID);
    glBindTexture(GL_TEXTURE_2D, TextureID);

    // Set up parameters
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, texWrapS);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, texWrapT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);

(* howabout
    // Zusatz : Anisotropisches Filtern
{!} if pAni then
{!}  begin

{!}  glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, @MaxAnisotropy);
{!}  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, MaxAnisotropy);
{!}  end
{!} else
{!}  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 0);
    // MipMaps generieren, wenn mglich im komprimierten S3TC-Format
{!} if pS3TC then
{!}  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_COMPRESSED_RGBA_S3TC_DXT1_EXT, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, PByte(rgbBits))
{!} else
{!}  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, PByte(rgbBits))
*)

{my}
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); { only first two can be used }
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); { all of the above can be used }
{!}  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, rgbBits);
  end;

end;

// Name       : Bind
// Desription : Binds this images texture ID to the current OpenGL texture
//              unit
// Parameters : None
procedure TGLBMP.Bind;
begin
  if TextureID > 0 then
    glBindTexture(GL_TEXTURE_2D, TextureID);
end;

// Name       : SetFilter
// Desription : Sets the texture minification and magnification filters
// Parameters :
//  min - texture minification filter
//  mag - texture magnification filter
procedure TGLBMP.SetFilter(min, mag : GLint);
begin
minfilter := min;
magfilter := mag;
exit;
  case min of
    GL_NEAREST : minFilter := GL_NEAREST;
    GL_LINEAR : minFilter := GL_LINEAR;
    GL_NEAREST_MIPMAP_NEAREST : minFilter := GL_NEAREST_MIPMAP_NEAREST;
    GL_LINEAR_MIPMAP_NEAREST : minFilter := GL_LINEAR_MIPMAP_NEAREST;
    GL_NEAREST_MIPMAP_LINEAR : minFilter := GL_NEAREST_MIPMAP_LINEAR;
    GL_LINEAR_MIPMAP_LINEAR : minFilter := GL_LINEAR_MIPMAP_LINEAR;
  end;

  case mag of
    GL_NEAREST : magFilter := GL_NEAREST;
    GL_LINEAR : magFilter := GL_LINEAR;
  end;
end;

// Name       : SetTextureWrap
// Desription : Sets the textures wrapping parameter
// Parameters :
//  s, t -  Wrap type for the s and t direction
procedure TGLBMP.SetTextureWrap(s, t: GLint);
begin
texWrapS := s;
texWrapT := t;
end;

// Name       : GetWidth
// Desription : Gets the current image width
// Returns    : image width
function TGLBMP.GetWidth : Integer;
begin
  Result := width;
end;

// Name       : GetHeight
// Desription : Gets the current image height
// Returns    : image height
function TGLBMP.GetHeight : Integer;
begin
  Result := height;
end;

// Name       : GetData
// Desription : Gets the current image data
// Returns    : pointer to image data
function TGLBMP.GetData : Pointer;
begin
  Result := PByte(rgbBits);
end;

// Name       : Scale8
// Purpose    : Scales the color value for a pixel in a 8bpp image
// Parameters :
//  colorValue - value to be scaled
//  colorBits - number of set bits in the color mask for this color
// Returns    : Scaled value (0-255)
function TGLBMP.Scale8(colorValue, colorBits : Integer) : Integer;
begin
  case colorBits of
    1: if Boolean(colorValue) then Result := 255 else Result := 0;
    2: Result := (colorValue shl 6) or (colorValue shl 4) or (colorValue shl 2)or colorValue;
    3: Result := (colorValue shl 5) or (colorValue shl 2) or (colorValue shr 1);
    4: Result := (colorValue shl 4) or colorValue;
    5: Result := (colorValue shl 3) or (colorValue shr 2);
    6: Result := (colorValue shl 2) or (colorValue shr 4);
    7: Result := (colorValue shl 1) or (colorValue shr 6);
    else Result := colorValue;
  end;
end;

function TGLBMP.GetLineWidth(bpp : Integer) : Integer;
begin
  Result := ((Width * bpp + 31) and - 32) shr 3;
end;

// Name       : GetRightShiftCount
// Purpose    : Calculates the amount needed to shift the color mask right
// Parameters :
//  dwVal - the color mask
// Returns    : Number of 0's before first set bit
function TGLBMP.GetRightShiftCount(dwVal : DWord): Integer;
var
  I : Integer;
begin
  for I := 0 to SizeOf(DWORD)*8 do begin
    if ((dwVal and 1) = 1) then begin
      Result := I;
      Exit;
    end;
    dwVal := dwVal shr 1;
  end;
  Result := -1;
end;

// Name       : GetNumSet
// Purpose    : Gets the number of set bits in the color mask
// Parameters :
//  dwVal - color mask
// Returns    : number of set bits
function TGLBMP.GetNumSet(dwVal : DWord): Integer;
var
  nCount : Integer;
begin
  nCount := 0;
  while (dwVal <> 0) do begin
    Inc(nCount, (dwVal and 1));
    dwVal := dwVal shr 1;
  end;

  Result := nCount;
end;

// Name       : SwapRB
// Desription : Swaps the red and blue channels for a 32bpp image
// Parameters : None
procedure TGLBMP.SwapRB;
var
  i, x, y : Integer;
  pixel : PGLRGBQUAD;
  shitpixel : PGLRGBQUAD;
  tmp : Byte;
  tempRGB:TGLRGBQUAD;
  picbits:PGLRGBQUAD;



begin

  // Loop through buffer swapping red and blue channels
  pixel := rgbBits;
  for x := 0 to Width - 1 do begin
    for y := 0 to Height - 1 do begin
      tmp := pixel^.blue;
      pixel^.blue := pixel^.red;
      pixel^.red := tmp;

      Inc(pixel);
    end;
  end;

end;

// Name       : ExpandTo32
// Desription : Takes the passed in image buffer and expands it from 24bpp into
//              the rgbBits 32bpp buffer.
// Parameters :
//  image - Pointer to the 24bpp image buffer
//  pad - Amount of padding at the end of each line in the 24bpp image
//  swapRB - Whether to swap the red and blue channels
// Returns    :
procedure TGLBMP.ExpandTo32(image : Pointer; pad : Integer; swapRBflag : Boolean);
var
  x, y : Integer;
  pixel24 : PGLRGBTRIPLE;
  pixel32 : PGLRGBQUAD;
begin

  // Loop through memory buffer copying it into 32bpp buffer and swapping
  // red/blue channels if neccessary
  pixel24 := image;
  pixel32 := rgbBits;
  for x := 0 to Width-1 do begin
    for y := 0 to Height-1 do begin

      if swapRBflag then begin
        pixel32^.red := pixel24^.blue;
        pixel32^.blue := pixel24^.red;
      end else begin
        pixel32^.red := pixel24^.red;
        pixel32^.blue := pixel24^.blue;
      end;
      pixel32^.green := pixel24^.green;

      Inc(pixel24);
      Inc(pixel32);
    end;
    pixel24 := Ptr(Integer(pixel24) + pad);
  end;
end;

// Name       : SetAlpha
// Desription : Sets the alpha channel for an image to one value
// Parameters :
//  alpha - desired alpha value
procedure TGLBMP.SetAlpha(alpha : Byte);
var
  I : Integer;
  pixel : PGLRGBQUAD;
begin
  pixel := rgbBits;
  for I := 0 to Width*Height - 1 do begin
    pixel^.alpha := alpha;
    Inc(pixel);
  end;
end;

// Name       : ClearMem
// Purpose    : Free's image memory
// Parameters : None
procedure TGLBMP.ClearMem;
begin
  if Assigned(rgbBits) then begin
    FreeMem(rgbBits,imageSize);{howabout}
    rgbBits := nil;
  end;
end;

// Name       : Destroy
// Purpose    : Frees used memory
// Parameters : None
procedure TGLBMP.Destroy;
begin
  // Free memory used by bitmap
  ClearMem();

  // Free the PNG libraries

  // Free the created texture ID, if needed
  if TextureID > 0 then
    glDeleteTextures(1, @TextureID);

end;

procedure TGLBMP.DestroyGL;
begin

  // Free the created texture ID, if needed
  if TextureID > 0 then
    glDeleteTextures(1, @TextureID);

end;


// Additions:
// SetPixel and AddPixel procedures
// by Georgy Moshkin

procedure TGLBMP.SetPixel(x,y:integer;r,g,b:byte);
var   pixel : PGLRGBQUAD;
begin
  Pixel := Pointer(Cardinal(rgbBits) + (x+1)*4 + Width*4*y - 4);
  Pixel^.Red:=r;
  Pixel^.Green:=g;
  Pixel^.blue:=b;
end;

procedure TGLBMP.AddPixel(x,y:integer;r,g,b:byte);
var   pixel : PGLRGBQUAD;
begin
  Pixel := Pointer(Cardinal(rgbBits) + (x+1)*4 + Width*4*y - 4);

  if Pixel^.Red+r>255 then
   Pixel^.Red:=255
  else
  Pixel^.Red:=Pixel^.Red+r;

  if Pixel^.Green+g>255 then
   Pixel^.Green:=255
  else
  Pixel^.Green:=Pixel^.Green+g;

  if Pixel^.Blue+b>255 then
   Pixel^.Blue:=255
  else
  Pixel^.Blue:=Pixel^.Blue+b;

end;


end.